/*
 * This file is part of the Score-P software (http://www.score-p.org)
 *
 * Copyright (c) 2009-2012,
 * RWTH Aachen University, Germany
 *
 * Copyright (c) 2009-2012,
 * Gesellschaft fuer numerische Simulation mbH Braunschweig, Germany
 *
 * Copyright (c) 2009-2012, 2014,
 * Technische Universitaet Dresden, Germany
 *
 * Copyright (c) 2009-2012,
 * University of Oregon, Eugene, USA
 *
 * Copyright (c) 2009-2012,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Copyright (c) 2009-2012,
 * German Research School for Simulation Sciences GmbH, Juelich/Aachen, Germany
 *
 * Copyright (c) 2009-2012,
 * Technische Universitaet Muenchen, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */

#ifndef CUBE_ERROR_CODES_H
#define CUBE_ERROR_CODES_H


/**
 * @file            CUBE_ErrorCodes.h
 * @ingroup         CUBE_Exception_module
 *
 * @brief           Error codes and error handling.
 *
 */

#include <errno.h>
#include <stdint.h>
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * This is the list of error codes for CUBE.
 */
typedef enum
{
    /** Special marker for error messages which indicates an deprecation. */
    CUBE_DEPRECATED    = -3,

    /** Special marker when the application will be aborted. */
    CUBE_ABORT         = -2,

    /** Special marker for error messages which are only warnings. */
    CUBE_WARNING       = -1,

    /** Operation successful */
    CUBE_SUCCESS       = 0,

    /** Invalid error code
     *
     * Should only be used internally and not as an actual error code.
     */
    CUBE_ERROR_INVALID = 1,

    /* These are the internal implementation of POSIX error codes. */
    /** The list of arguments is to long */
    CUBE_ERROR_E2BIG,
    /** Not enough rights */
    CUBE_ERROR_EACCES,
    /** Address is not available */
    CUBE_ERROR_EADDRNOTAVAIL,
    /** Address family is not supported */
    CUBE_ERROR_EAFNOSUPPORT,
    /** Resource temporaly not available */
    CUBE_ERROR_EAGAIN,
    /** Connection is already processed */
    CUBE_ERROR_EALREADY,
    /** Invalid file pointer */
    CUBE_ERROR_EBADF,
    /** Invalid message */
    CUBE_ERROR_EBADMSG,
    /** Resource or device is busy */
    CUBE_ERROR_EBUSY,
    /** Operation was aborted */
    CUBE_ERROR_ECANCELED,
    /** No child process available */
    CUBE_ERROR_ECHILD,
    /** Connection was refused */
    CUBE_ERROR_ECONNREFUSED,
    /** Connection was reset */
    CUBE_ERROR_ECONNRESET,
    /** Resolved deadlock */
    CUBE_ERROR_EDEADLK,
    /** Destination address was expected */
    CUBE_ERROR_EDESTADDRREQ,
    /** Domain error */
    CUBE_ERROR_EDOM,
    /** Reserved */
    CUBE_ERROR_EDQUOT,
    /** File does already exist */
    CUBE_ERROR_EEXIST,
    /** Invalid Address */
    CUBE_ERROR_EFAULT,
    /** File is to big */
    CUBE_ERROR_EFBIG,
    /** Operation is work in progress */
    CUBE_ERROR_EINPROGRESS,
    /** Interuption of an operating system call */
    CUBE_ERROR_EINTR,
    /** Invalid argument */
    CUBE_ERROR_EINVAL,
    /** Generic I/O error */
    CUBE_ERROR_EIO,
    /** Socket is already connected */
    CUBE_ERROR_EISCONN,
    /** Target is a directory */
    CUBE_ERROR_EISDIR,
    /** To many layers of symbolic links */
    CUBE_ERROR_ELOOP,
    /** To many opened files */
    CUBE_ERROR_EMFILE,
    /** To many links */
    CUBE_ERROR_EMLINK,
    /** Message buffer is to small */
    CUBE_ERROR_EMSGSIZE,
    /** Reserved */
    CUBE_ERROR_EMULTIHOP,
    /** Filename is to long */
    CUBE_ERROR_ENAMETOOLONG,
    /** Network is down */
    CUBE_ERROR_ENETDOWN,
    /** Connection was reset from the network */
    CUBE_ERROR_ENETRESET,
    /** Network is not reachable */
    CUBE_ERROR_ENETUNREACH,
    /** To much opened files */
    CUBE_ERROR_ENFILE,
    /** No buffer space available */
    CUBE_ERROR_ENOBUFS,
    /** No more data left in the queue */
    CUBE_ERROR_ENODATA,
    /** This device does not support this function */
    CUBE_ERROR_ENODEV,
    /** File or Directory does not exist */
    CUBE_ERROR_ENOENT,
    /** Cannot execute binary */
    CUBE_ERROR_ENOEXEC,
    /** Locking failed */
    CUBE_ERROR_ENOLCK,
    /** Reserved */
    CUBE_ERROR_ENOLINK,
    /** Not enough main memory available */
    CUBE_ERROR_ENOMEM,
    /** Message has not the expected type */
    CUBE_ERROR_ENOMSG,
    /** This protocol is not available */
    CUBE_ERROR_ENOPROTOOPT,
    /** No space left on device */
    CUBE_ERROR_ENOSPC,
    /** No stream available */
    CUBE_ERROR_ENOSR,
    /** This is not a stream */
    CUBE_ERROR_ENOSTR,
    /** Requested function is not implemented */
    CUBE_ERROR_ENOSYS,
    /** Socket is not connected */
    CUBE_ERROR_ENOTCONN,
    /** This is not an directory */
    CUBE_ERROR_ENOTDIR,
    /** This directory is not empty */
    CUBE_ERROR_ENOTEMPTY,
    /** No socket */
    CUBE_ERROR_ENOTSOCK,
    /** This operation is not supported */
    CUBE_ERROR_ENOTSUP,
    /** This IOCTL is not supported by the device */
    CUBE_ERROR_ENOTTY,
    /** Device is not yet configured */
    CUBE_ERROR_ENXIO,
    /** Operation is not supported by this socket */
    CUBE_ERROR_EOPNOTSUPP,
    /** Value is to long for the datatype */
    CUBE_ERROR_EOVERFLOW,
    /** Operation is not permitted */
    CUBE_ERROR_EPERM,
    /** Broken pipe */
    CUBE_ERROR_EPIPE,
    /** Protocoll error */
    CUBE_ERROR_EPROTO,
    /** Protocoll is not supported */
    CUBE_ERROR_EPROTONOSUPPORT,
    /** Wrong protocoll type for this socket */
    CUBE_ERROR_EPROTOTYPE,
    /** Value is out of range */
    CUBE_ERROR_ERANGE,
    /** Filesystem is read only */
    CUBE_ERROR_EROFS,
    /** This seek is not allowed */
    CUBE_ERROR_ESPIPE,
    /** No matching process found */
    CUBE_ERROR_ESRCH,
    /** Reserved */
    CUBE_ERROR_ESTALE,
    /** Timout in file stream or IOCTL */
    CUBE_ERROR_ETIME,
    /** Connection timed out */
    CUBE_ERROR_ETIMEDOUT,
    /** File couldn't be executed while it is opened */
    CUBE_ERROR_ETXTBSY,
    /** Operation would be blocking */
    CUBE_ERROR_EWOULDBLOCK,
    /** Invalid link between devices */
    CUBE_ERROR_EXDEV,

    /* These are the error codes specific to the Cube package */

    /** Function call not allowed in current state */
    CUBE_ERROR_INVALID_CALL,
    /** Parameter value out of range */
    CUBE_ERROR_INVALID_ARGUMENT,
    /** Invalid or inconsistent record data */
    CUBE_ERROR_INVALID_DATA,
    /** The given size cannot be used */
    CUBE_ERROR_INVALID_SIZE_GIVEN,
    /** The given type is not known */
    CUBE_ERROR_UNKNOWN_TYPE,
    /** This could not be done with the given memory */
    CUBE_ERROR_MEM_FAULT,
    /** Memory allocation failed */
    CUBE_ERROR_MEM_ALLOC_FAILED,
    /** Index out of bounds */
    CUBE_ERROR_INDEX_OUT_OF_BOUNDS,
    /** Invalid source code line number */
    CUBE_ERROR_INVALID_LINENO,
    /** End of buffer/file reached */
    CUBE_ERROR_END_OF_BUFFER,
    /** Unable to open file */
    CUBE_ERROR_FILE_CAN_NOT_OPEN,
    /** Missing library support for requested compression mode */
    CUBE_ERROR_FILE_COMPRESSION_NOT_SUPPORTED,
    /** Invalid file operation */
    CUBE_ERROR_FILE_INTERACTION
} CUBE_ErrorCode;


/**
 * Returns the name of an error code.
 *
 * @param errorCode : Error Code
 *
 * @returns Returns the name of a known error code, and "INVALID_ERROR" for
 *          invalid or unknown error IDs.
 * @ingroup CUBE_Exception_module
 */
const char*
CUBE_Error_GetName( CUBE_ErrorCode errorCode );


/**
 * Returns the description of an error code.
 *
 * @param errorCode : Error Code
 *
 * @returns Returns the description of a known error code.
 * @ingroup CUBE_Exception_module
 */
const char*
CUBE_Error_GetDescription( CUBE_ErrorCode errorCode );


/**
 * Signature of error handler callback functions. The error handler can be set
 * with @ref CUBE_Error_RegisterCallback.
 *
 * @param userData        : Data passed to this function as given by the registry call.
 * @param file            : Name of the source-code file where the error appeared
 * @param line            : Line number in the source-code where the error appeared
 * @param function        : Name of the function where the error appeared
 * @param errorCode       : Error Code
 * @param msgFormatString : Format string like it is used at the printf family.
 * @param va              : Variable argument list
 *
 * @returns Should return the errorCode
 */
typedef CUBE_ErrorCode
( *CUBE_ErrorCallback )( void*          userData,
                         const char*    file,
                         uint64_t       line,
                         const char*    function,
                         CUBE_ErrorCode errorCode,
                         const char*    msgFormatString,
                         va_list        va );


/**
 * Register a programmers callback function for error handling.
 *
 * @param errorCallbackIn : Fucntion will be called instead of printing a
 *                          default message to standard error.
 * @param userData :        Data pointer passed to the callback.
 *
 * @returns Function pointer to the former error handling function.
 * @ingroup CUBE_Exception_module
 *
 */
CUBE_ErrorCallback
CUBE_Error_RegisterCallback( CUBE_ErrorCallback errorCallbackIn,
                             void*              userData );


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* CUBE_ERROR_CODES_H */
